package interpreter

import org.scalacheck.*
import Gen.*
import Arbitrary.arbitrary

object ExprGen:
  import NumOp.*
  import PairOp.*
  import Expr.*

  enum ValueType:
    case NumT
    case PairT(v1: ValueType, v2: ValueType)

  import ValueType.*

  def randomType(size: Int = 3): Gen[ValueType] =
    if size <= 0 then const(NumT)
    else
      frequency(
        (1, const(NumT)),
        (
          1,
          for
            v1 <- randomType(size - 1)
            v2 <- randomType(size - 1)
          yield PairT(v1, v2)
        )
      )

  def genName(using env: Map[String, ValueType]): Gen[String] =
    val genStr =
      for
        c <- choose('a', 'z')
        i <- choose(0, 99)
        d <- choose(0, 1)
      yield
        def recur(j: Int): String =
          val n = s"$c$j"
          if env.contains(n) then recur(j * 2 + d)
          else n
        recur(i)

    genStr.filter(!env.contains(_))

  def genNumConst: Gen[Num] =
    for n <- choose(-100, 100)
    yield Num(n)

  def genNumExpr(size: Int)(using env: Map[String, ValueType]): Gen[NumExpr] =
    for
      // without Div, there will be no division by zero error in the generated programs
      op <- oneOf(Add, Sub, Mul) // , Div)
      e1 <- genExpr(size - 1, NumT)
      e2 <- genExpr(size - 1, NumT)
    yield NumExpr(op, e1, e2)

  def genPair(size: Int, tp1: ValueType, tp2: ValueType)(using
      env: Map[String, ValueType]
  ): Gen[Pair] =
    for
      e1 <- genExpr(size - 1, tp1)
      e2 <- genExpr(size - 1, tp2)
    yield Pair(e1, e2)

  def genPairExpr(size: Int, tp: ValueType)(using
      env: Map[String, ValueType]
  ): Gen[PairExpr] =
    for
      op <- oneOf(First, Second)
      rdT <- randomType(size - 2)
      e <- op match
        case First  => genExpr(size - 1, PairT(tp, rdT))
        case Second => genExpr(size - 1, PairT(rdT, tp))
    yield PairExpr(op, e)

  def genIsPair(size: Int)(using env: Map[String, ValueType]): Gen[IsPair] =
    for
      rdT <- randomType(size - 1)
      e <- genExpr(size - 1, rdT)
    yield IsPair(e)

  def genVar(using env: Map[String, ValueType]): Gen[Expr] =
    lzy { oneOf(env.keys).map(Var(_)) }

  def genLet(size: Int, tp: ValueType)(using
      env: Map[String, ValueType]
  ): Gen[Let] =
    for
      name <- genName
      rdT <- randomType(size - 1)
      e <- genExpr(size - 1, rdT)
      body <- genExpr(size - 1, tp)(using env + (name -> rdT))
    yield Let(name, e, body)

  def genIfEqZero(size: Int, tp: ValueType)(using
      env: Map[String, ValueType]
  ): Gen[If] =
    for
      cond1 <- genExpr(size - 1, NumT)
      cond2 <-
        genName.map(tmp => Let(tmp, cond1, NumExpr(Sub, Var(tmp), Var(tmp))))
      cond <- oneOf(cond1, cond2)
      thenb <- genExpr(size - 1, tp)
      elseb <- genExpr(size - 1, tp)
    yield If(cond, thenb, elseb)

  def genExpr(size: Int, tp: ValueType)(using
      env: Map[String, ValueType]
  ): Gen[Expr] = lzy {
    // println(s"genExpr($size, $tp)")
    val envWithTp = env.filter(_._2 == tp)
    tp match
      case NumT =>
        frequency(
          (1, genNumConst),
          (envWithTp.size, genVar(using envWithTp)),
          (if size > 0 then 1 else 0, genNumExpr(size)),
          (if size > 0 then 1 else 0, genPairExpr(size, tp)),
          (if size > 0 then 1 else 0, genIsPair(size)),
          (if size > 0 then 1 else 0, genLet(size, tp)),
          (if size > 0 then 1 else 0, genIfEqZero(size, tp))
        )
      case PairT(tp1, tp2) =>
        frequency(
          (1, genPair(size, tp1, tp2)),
          (envWithTp.size, genVar(using envWithTp)),
          (if size > 0 then 1 else 0, genPairExpr(size, tp)),
          (if size > 0 then 1 else 0, genLet(size, tp)),
          (if size > 0 then 1 else 0, genIfEqZero(size, tp))
        )
  }

  def getRandExpr(size: Int): Expr =
    val tp = randomType(size).sample.get
    genExpr(size, tp)(using Map()).sample.get

object ExprHelpers:
  import NumOp.*
  import PairOp.*
  import Expr.*

  val simple1 =
    Let(
      "x",
      Num(1),
      Let("y", NumExpr(Add, Var("x"), Num(1)), Pair(Var("x"), Var("y")))
    )

  val simple2 =
    Let(
      "x",
      Num(1),
      Let(
        "y",
        Num(2),
        NumExpr(
          Add,
          PairExpr(First, Pair(Var("x"), Num(4))),
          PairExpr(Second, Pair(Num(3), Var("y")))
        )
      )
    )

  val falseConditional =
    Let(
      "x",
      Num(1),
      Let(
        "y",
        Pair(Num(2), Var("x")),
        If(
          NumExpr(Sub, Num(1), PairExpr(Second, Var("y"))),
          Var("y"),
          Pair(Num(0), Var("x"))
        )
      )
    )

  val trueConditional =
    If(NumExpr(Add, Num(1), Num(4)), Pair(Num(0), Num(0)), Num(2))

  val divByZero = Let("x", NumExpr(Div, Num(1), Num(0)), Var("x"))

  val divByZero2 =
    Let("x", NumExpr(Div, Num(1), Num(0)), Num(1))

  val unboundVar = Var("z")

  val numOpOnNonNum = NumExpr(Add, Num(1), Pair(Num(1), Num(2)))

  val pairOpOnNonPair = PairExpr(First, Num(1))

  val ifCondNotNum = If(Pair(Num(1), Num(2)), Num(2), Num(1))
